<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

namespace block_zoomonline;
/**
 * Process Zoom meetings task.
 *
 * @package    block_zoomonline
 * @copyright  2024 Ciaran Mac Donncha
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class ZoomHelper {

    /**
     * @var array|null Stores the cached access token and its timestamp.
     */
    private static $cachedtoken = null;

    /**
     * @var int The expiry time for the access token in seconds (set to 3500 seconds or ~58 minutes).
     */
    private static $tokenexpirytime = 3500; // Token is valid for 1 hour, subtract a buffer.

    /**
     * Create a new Zoom meeting.
     *
     * @param string $hostemail The host email for the meeting.
     * @param string $name The name of the meeting.
     * @return string|null The Zoom meeting ID.
     * @throws moodle_exception
     */
    public static function createmeeting($hostemail, $name) {
        $accesskey = self::getcurrentaccesskey();
        $postdata = [
                "topic" => $name,
                "type" => "3",
                "timezone" => "Europe/Dublin",
                "schedule_for" => $hostemail,
                "agenda" => $name,
                "settings" => [
                        "use_pmi" => false,
                        "auto_recording" => "cloud",
                ],
        ];

        $postfields = json_encode($postdata);

        do {
            $url = "users/$hostemail/meetings";
            $response = self::executezoomapirequest($url, 'POST', $accesskey, $postfields);

            // If token has expired, generate a new one and retry.
            if (isset($response['error']) && $response['error'] == '401') {
                $accesskey = self::generatenewtoken();
            } else {
                // Break loop if success or any other error occurs.
                break;
            }
        } while (true);

        if (isset($response['id'])) {
            return $response['id'];
        }

        debugging('Error in Zoom meeting creation response: ' . json_encode($response), DEBUG_DEVELOPER);
        throw new moodle_exception('zoomapierror', 'block_zoomonline', '', null, 'Error in Zoom API response');
    }

    /**
     * Get the current Zoom access token, regenerating if expired.
     *
     * @return string The current access token.
     * @throws moodle_exception
     */
    public static function getcurrentaccesskey() {
        global $DB, $CFG;

        // Check if we already have a cached token within the execution context.
        if (self::$cachedtoken !== null && (time() - self::$cachedtoken['timestamp']) < self::$tokenexpirytime) {
            return self::$cachedtoken['access_token'];
        }

        // Fetch the latest access token from the database.
        $sqlselect = "SELECT id, current, access_token FROM {block_zoomonline_token} ORDER BY id DESC LIMIT 1";
        $record = $DB->get_record_sql($sqlselect);

        // If the token is expired or not found, generate a new one.
        if (!$record || (time() - $record->current > self::$tokenexpirytime)) {
            $newtoken = self::generatenewtoken();
            self::$cachedtoken = ['access_token' => $newtoken, 'timestamp' => time()];
            return $newtoken;
        }

        // Cache the token for the duration of the request and return it.
        self::$cachedtoken = ['access_token' => $record->access_token, 'timestamp' => $record->current];
        return $record->access_token;
    }

    /**
     * Generate a new Zoom access token.
     *
     * @return string The new access token.
     * @throws moodle_exception
     */
    public static function generatenewtoken() {
        global $DB, $CFG;
        require_once($CFG->dirroot . '/local/guzzle/extlib/vendor/autoload.php');

        [$clientid, $clientsecret, $accountid] = self::getzoomcredentials();

        try {
            // Cleanup old tokens (keep last 5 tokens for audit purposes).
            $sql = "DELETE FROM {block_zoomonline_token}
              WHERE id NOT IN (
                  SELECT id FROM (
                      SELECT id FROM {block_zoomonline_token}
                      ORDER BY id DESC LIMIT 5
                  ) as latest_tokens
              )";
            $DB->execute($sql);

            // Request a new access token using Guzzle HTTP client.
            $client = new \GuzzleHttp\Client(['base_uri' => 'https://zoom.us']);
            $response = $client->request('POST', '/oauth/token', [
                    'query' => [
                            'grant_type' => 'account_credentials',
                            'account_id' => $accountid,
                    ],
                    'headers' => [
                            'Authorization' => 'Basic ' . base64_encode("$clientid:$clientsecret"),
                            'Content-Type' => 'application/json',
                    ],
            ]);

            $contents = json_decode($response->getBody()->getContents());
            $accesstoken = $contents->access_token ?? null;
            $currenttime = time();

            if (!empty($accesstoken)) {
                // Store the token in the database.
                $record = new \stdClass();
                $record->access_token = $accesstoken;
                $record->current = $currenttime;

                $DB->insert_record('block_zoomonline_token', $record);

                // Update the cached token with the new token.
                self::$cachedtoken = ['access_token' => $accesstoken, 'timestamp' => $currenttime];

                return $accesstoken;
            } else {
                throw new moodle_exception('invalidresponse', 'block_zoomonline', '', null, 'Invalid token response from Zoom API');
            }
        } catch (\Exception $e) {
            debugging('Zoom API token generation failed: ' . $e->getMessage(), DEBUG_DEVELOPER);
            throw new moodle_exception('tokenerror', 'block_zoomonline', '', null, $e->getMessage());
        }

    }

    /**
     * Retrieves the Zoom API credentials from the plugin configuration.
     *
     * @return array An array containing the client ID, client secret, and account ID.
     * @throws moodle_exception If any of the required credentials are missing.
     */
    private static function getzoomcredentials() {
        $clientid = get_config('block_zoomonline', 'client_id');
        $clientsecret = get_config('block_zoomonline', 'client_secret');
        $accountid = get_config('block_zoomonline', 'account_id');

        if (empty($clientid) || empty($clientsecret) || empty($accountid)) {
            self::handleapierror('Missing Zoom API credentials');
        }

        return [$clientid, $clientsecret, $accountid];
    }

    /**
     * Handles API errors by logging them and optionally throwing an exception.
     *
     * @param string $errormessage The error message to log and potentially display.
     * @param array $response The full API response, if available.
     * @param bool $throwexception Whether to throw a moodle_exception or not.
     * @throws moodle_exception If $throwexception is true.
     */
    public static function handleapierror($errormessage, $response = [], $throwexception = true) {
        // Log the error message to Moodle's debug log.
        debugging('Zoom API Error: ' . $errormessage, DEBUG_DEVELOPER);

        // Optionally, log more detailed information from the API response if available.
        if (!empty($response)) {
            debugging('Zoom API Response: ' . json_encode($response), DEBUG_DEVELOPER);
        }

        // You can notify administrators here if necessary (optional):
        // For example, you can send an email to admins when a critical error occurs.

        // Prepare a user-friendly error message for Moodle.
        $friendlyerrormessage = get_string('zoomapierror', 'block_zoomonline', $errormessage);

        // Optionally, display an error message to the user in the Moodle interface.
        if ($throwexception) {
            throw new moodle_exception('zoomapierror', 'block_zoomonline', '', null, $friendlyerrormessage);
        }
    }

    /**
     * Set cURL options and execute the Zoom API request.
     *
     * @param string $url API URL.
     * @param string $method HTTP method (GET, POST, etc.).
     * @param string $accesskey Access token.
     * @param mixed|null $postoptions POST data if applicable.
     * @return array The decoded response from the Zoom API.
     * @throws moodle_exception
     */
    private static function executezoomapirequest($url, $method, $accesskey, $postoptions = null) {
        $curl = curl_init();
        self::setcurloptions($url, $curl, $method, $accesskey, $postoptions);
        $response = curl_exec($curl);
        $err = curl_error($curl);
        curl_close($curl);

        if ($err) {
            throw new moodle_exception('zoomapierror', 'block_zoomonline', '', null, $err);
        }

        return json_decode($response, true);
    }

    /**
     * Set cURL options for a Zoom API request.
     *
     * @param string $url The API endpoint URL.
     * @param resource $curl The cURL resource.
     * @param string $method The HTTP method.
     * @param string $accesskey The Zoom access token.
     * @param mixed|null $postoptions POST data if applicable.
     */
    private static function setcurloptions($url, $curl, $method, $accesskey, $postoptions = null) {
        $curloptions = [
                CURLOPT_URL => "https://api.zoom.us/v2/$url",
                CURLOPT_RETURNTRANSFER => true,
                CURLOPT_ENCODING => "",
                CURLOPT_MAXREDIRS => 10,
                CURLOPT_TIMEOUT => 30,
                CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
                CURLOPT_CUSTOMREQUEST => $method,
                CURLOPT_HTTPHEADER => [
                        "Authorization: Bearer $accesskey",
                        "Content-Type: application/json",
                ],
        ];

        if ($postoptions) {
            $curloptions[CURLOPT_POSTFIELDS] = $postoptions;
        }

        curl_setopt_array($curl, $curloptions);
    }

    /**
     * Check if a Zoom meeting exists.
     *
     * @param string $meetingid The Zoom meeting ID.
     * @param string $hostemail The host email for the meeting.
     * @return bool True if the meeting exists, false otherwise.
     */
    public static function checkmeetingexists($meetingid, $hostemail) {
        $accesskey = self::getcurrentaccesskey();
        $response = self::executezoomapirequest("meetings/$meetingid", 'GET', $accesskey);

        if ($response && isset($response['id']) && isset($response['host_email'])) {
            return (trim($response['id']) === trim($meetingid) &&
                    strtolower(trim($response['host_email'])) === strtolower(trim($hostemail)));
        }

        return false;
    }

    /**
     * Get meeting data from Zoom API.
     *
     * @param string $meetingid
     * @return array|null Meeting data or null if error.
     */
    public static function getmeetingdata($meetingid) {

        $accesskey = self::getcurrentaccesskey();
        $response = self::executezoomapirequest("report/meetings/$meetingid", 'GET', $accesskey);
        // Check if the response is already an array.
        if (is_array($response)) {
            // Return the array directly since it’s already decoded.
            return $response;
        }
        // Otherwise, decode the JSON string.
        return $response ? json_decode($response, true) : null;

    }

    /**
     * Retrieves the participants of a Zoom meeting.
     *
     * @param string $meetingid The ID of the Zoom meeting.
     * @param string $nextpagetoken The token for the next page of results, if any.
     * @return array The participant data from the Zoom API.
     */
    public static function getparticipants($meetingid, $nextpagetoken = "") {
        $accesskey = self::getcurrentaccesskey();

        // Construct the URL with the next_page_token if it exists.
        $url = "report/meetings/$meetingid/participants";

        if (!empty($nextpagetoken)) {
            $url .= "?next_page_token=" . urlencode($nextpagetoken); // Properly append next_page_token.
        }

        // Execute the Zoom API request.
        $response = self::executezoomapirequest($url, 'GET', $accesskey);

        return $response;
    }

    /**
     * Fetch Zoom meeting recordings.
     *
     * @param string $meetingid The Zoom meeting ID.
     * @return array The list of recording files.
     */
    public static function getmeetingrecordings($meetingid) {
        $accesskey = self::getcurrentaccesskey();
        $response = self::executezoomapirequest("meetings/$meetingid/recordings", 'GET', $accesskey);

        return $response['recording_files'] ?? [];
    }

    /**
     * Delete Zoom meeting recording.
     *
     * @param string $meetingid The Zoom meeting ID.
     * @param string $videoid The recording video ID.
     */
    public static function deletezoomrecording($meetingid, $videoid) {
        $accesskey = self::getcurrentaccesskey();
        self::executezoomapirequest("meetings/$meetingid/recordings/$videoid?action=trash", 'DELETE', $accesskey);
    }

    /**
     * Retrieves the current Zoom meeting status.
     *
     * @param string $meetingid The ID of the Zoom meeting.
     * @return array|null The meeting status data or null if an error occurred.
     */
    public static function getmeetingstatus($meetingid) {
        $accesskey = self::getcurrentaccesskey();
        $response = self::executezoomapirequest("/meetings/$meetingid", 'GET', $accesskey);

        return $response ?? null;
    }

    /**
     * Retrieves metrics for a specific Zoom meeting.
     *
     * @param string $meetingid The ID of the Zoom meeting.
     * @return array|null The meeting metrics data or null if an error occurred.
     */
    public static function getmeetingmetrics($meetingid) {
        $accesskey = self::getcurrentaccesskey();
        $response = self::executezoomapirequest("/metrics/meetings/$meetingid", 'GET', $accesskey);

        return $response ?? null;
    }

    /**
     * Downloads a Zoom video file from a given URL.
     *
     * @param string $url The URL of the video to download.
     * @param string $accesskey The Zoom access token for authentication.
     * @param string $videoid The unique identifier of the video.
     * @return string|null The path to the downloaded file, or null if the download failed.
     */
    public static function downloadvideo($url, $accesskey, $videoid) {
        global $CFG;

        $localfolder = get_config('block_zoomonline', 'localfolder');

        if ($localfolder != '') {
            $localdir = $CFG->dataroot . '/' . $localfolder;

            if (!file_exists($localdir)) {
                mkdir($localdir, 0777, true);
            }
            $tempdir = $localdir;
        } else {
            $tempdir = $localfolder ?: sys_get_temp_dir();
        }

        $tempfilepath = $tempdir . DIRECTORY_SEPARATOR . $videoid . ".mp4";
        if (!is_dir($tempdir) || !is_writable($tempdir)) {
            echo "Directory is not writable or does not exist: $tempdir";
            return null;
        }

        $fp = fopen($tempfilepath, 'wb');

        if ($fp === false) {
            echo "Unable to open file for writing: $tempfilepath";
            return null;
        }

        $ch = curl_init();

        $fp = fopen($tempfilepath, 'wb');
        $params = ['access_token' => $accesskey];
        $curl = $url . '?' . http_build_query($params);

        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_REFERER, $url);
        curl_setopt($ch, CURLOPT_URL, $curl);
        curl_setopt($ch, CURLOPT_FILE, $fp);
        curl_setopt($ch, CURLOPT_HEADER, 0);
        curl_exec($ch);
        curl_close($ch);
        fclose($fp);

        if (file_exists($tempfilepath) && filesize($tempfilepath) > 0) {
            return $tempfilepath;
        } else {
            echo "Downloaded file is empty.";
            return null;
        }
    }
}
